main = print $ toString $ simplify test

data Expr =
      Const Integer
    | Var String
    | Add Expr Expr
    | Mult Expr Expr
    | Nothing
    deriving (Show, Eq)


-----------------------------------------------------------------------------

diff :: Expr -> Expr -> Expr
diff (Const _) (Var z) = Const 0
diff (Var x) (Var z) =
    if x == z
      then Const 1
      else Var x
                        
diff (Add x y) (Var z) =
    Add (diff x (Var z)) (diff y (Var z))
  
diff (Mult x y) (Var z) =
    Add (Mult (diff x (Var z)) y) (Mult x (diff y (Var z)))

------------------------------------------------------------------------------

simplify :: Expr -> Expr
simplify (Var x) = Var x
simplify (Const x) = Const x

simplify (Mult (Const x) (Const y)) = Const (x * y)
simplify (Mult x y)
    | x == Const 1 = simplify y
    | y == Const 1 = simplify x
    | x == Const 0 = Const 0
    | y == Const 0 = Const 0
    | z == t       = simplify $ Mult (Const 2) z
    | otherwise    = simplify $ Mult (simplify x) (simplify y)
  where
    z = simplify x
    t = simplify y

simplify (Add (Const x) (Const y)) = Const (x + y)
simplify (Add x y)
    | x == Const 0 = Const 0
    | y == Const 0 = Const 0
    | z == t       = simplify $ Mult (Const 2) z
    | otherwise    = simplify $ Add (simplify x) (simplify y)
  where
    z = simplify x
    t = simplify y

------------------------------------------------------------------------------

toString :: Expr -> String
toString (Var x) = x
toString (Const x) = show x
toString (Add x y) = toString x ++ "+" ++ toString y

toString (Mult (Add x y) (Add z t)) =
    toStringHelper (Add x y) ++ "*" ++ toStringHelper (Add z t)

toString (Mult x (Add y z)) =
    toString x ++ "*" ++ toStringHelper (Add y z)

toString (Mult (Add x y) z) =
    toString (Mult z (Add x y))

toString (Mult x y) = toString x ++ "*" ++ toString y

toStringHelper :: Expr -> String
toStringHelper x = "(" ++ toString x ++ ")"

------------------------------------------------------------------------------

eval :: Expr -> [(String, Integer)] -> Expr
eval (Add x y) z = simplify $ Add (eval x z) (eval y z)
eval (Mult x y) z = simplify $ Mult (eval x z) (eval y z)
eval (Const x) z = (Const x)

eval (Var x) z
    | t == Main.Nothing = (Var x)
    | otherwise         = t
  where t = find x z

find :: String -> [(String, Integer)] -> Expr
find x y
    | t == []   = Main.Nothing
    | otherwise = Const (snd $ head t)
  where t = dropWhile (\(z, _) -> x /= z) y

------------------------------------------------------------------------------

parseExpr ::  String -> Expr
parseExpr _ = Const 0

------------------------------------------------------------------------------

test = diff (Add (Var "x") (Mult (Var "x") (Var "x"))) (Var "x")

